/*
 * LITpro_basic.i --
 *
 *	Implementation of basic routines needed by simple user model functions
 *	in LITpro software.
 *
 *-----------------------------------------------------------------------------
 *
 *      Copyright (C) 2006, the LITpro Team.
 *
 *	This file is part of LITpro.
 *
 *	LITpro is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	LITpro is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with LITpro; if not, write to the Free Software Foundation,
 *	Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 *
 *-----------------------------------------------------------------------------
 *
 *	$Id: LITpro_basic.i,v 1.19 2008-04-01 19:56:13 bosc Exp $
 *	$Log: LITpro_basic.i,v $
 *	Revision 1.19  2008-04-01 19:56:13  bosc
 *	ring function re-written
 *
 *	Revision 1.18  2008-03-18 09:43:39  michel
 *	- typo error corrected in lpb_elong_ring()
 *	- corrected bug in percentages _lpb_test_function()
 *
 *	Revision 1.17  2008-03-17 16:08:01  michel
 *	- (forgot comments on previous commit. Redo commit with no other
 *	  modifications)
 *	- lpb_function() now handle VMIN and VMAX fields. VMIN and VMAX fields
 *	  have been filled for all the functions.
 *	- major review of the geometric functions (limb_* not modified)
 *	- lbp_ring should be more stable on singular points.
 *	- review of all the error messages.
 *	- _lpb_test_functions() added. Check functions on singular points.
 *
 *	Revision 1.16  2008-03-17 15:59:18  michel
 *	- major review of the geometric functions (limb_* not modified)
 *
 *	Revision 1.15  2008-03-13 14:32:02  bosc
 *	change into ring function and return to _LPB_DBL_MIN in circular functions
 *
 *	Revision 1.12  2008-02-22 13:09:04  jmmc
 *	Fixed missing radius_ratio of elong_ring's UNITS
 *
 *	Revision 1.11  2008-02-21 21:00:14  bosc
 *	*** empty log message ***
 *
 *	Revision 1.10  2007-11-19 16:38:17  michel
 *	- fixed underflow in lpb_disk (w * J1(2r)/r) -> (J1(2r)/r) * w
 *
 *	Revision 1.9  2007-11-19 16:17:06  cbechet
 *	Fixed two bug in the fitter, about new_active array and rescaled step. Added documentation and unified NPARAM hash member. N_PARAM is no longer used.
 *
 *	Revision 1.8  2007-03-09 12:20:16  michel
 *	- function lpb_functions() added.
 *	
 *	Revision 1.7  2006/12/07 15:37:23  michel
 *	- added variable _LPB_DEG2RAD and few comments.
 *	
 *	Revision 1.6  2006/06/11 22:53:10  bosc
 *	test on the variables added; lpb_ring corrected;
 *	elongated functions modified (ratio b/a instead of b and a)
 *	
 *	Revision 1.5  2006/06/08 16:07:43  bosc
 *	center-to-limb darkening functions added.
 *	
 *	Revision 1.4  2006/06/06 21:57:21  bosc
 *	change of the name of the functions, more homogeneous.
 *	lpb_punct,disk,circle,gaussian,ring and lpb_elong_disk,gaussian,ring tested
 *	with a plot of the result and of its fft (=object map)
 *	
 *	Revision 1.3  2006/06/02 11:31:58  eric
 *	changed prefix oi?_ into lp?_
 *	
 *	Revision 1.2  2006/06/02 09:42:27  michel
 *	- First functions (lpb_punct, lpb_unifdisk, lpb_gaussian) made.
 *	
 *	Revision 1.1.1.1  2006/05/31 08:33:15  michel
 *	Yorick Optical-interferometry General Algorithms
 *	
 *-----------------------------------------------------------------------------
 */

/* TODO
 * - update limp_* functions.
 * 
 */

local _LPB_DBL_MIN; 
/* DOCUMENT _LPB_DBL_MIN
    Smallest (in magnitude) useable floating value, consistent with
    there being no leading zeros in the mantissa. Known also as
    minimum normalized positive floating-point number.
*/
_LPB_DBL_MIN= machine_constant("DBL_MIN");

local _LPB_DBL_EPS;
/* DOCUMENT _LPB_DBL_EPS
   Smallest (in magnitude) positive floating-point number that, added 
   to 1., gives something different from 1.  Loosely referred to as the 
   "floating-point precision".
*/
_LPB_DBL_EPS = machine_constant("DBL_EPSILON");


local _LPB_PI, _LPB_DEG2RAD, _LPB_MAS2RAD;
/* DOCUMENT   _LPB_PI = value of the variable PI, to avoid any corruption
    -or- _LPB_DEG2RAD = degree to radian conversion factor
    -or- _LPB_MAS2RAD = milliarcsecond to radian conversion factor
*/
_LPB_PI = 3.141592653589793238462643383279503;
_LPB_DEG2RAD = _LPB_PI/180.0;
_LPB_MAS2RAD = _LPB_PI/180.0/3600/1000;


local _LPB_FILE;
_LPB_FILE = current_include();  // Name of current file, used by lpb_functions.
 
func lpb_functions(void)
/* DOCUMENT lpb_functions
            h = lpb_functions()
    List the dictionary of the available standard modeling functions in LITpro
    library. When called as a subroutine (first form), the list of functions
    is printed on the terminal. When called as a function (second form),
    a hash table is returned, containing various informations on each
    function (list of arguments, units for each of them, help, ...).
    This second form is particularly useful to let the GUI know of this
    information.
    
    For a function to be selected in the dictionnary, it must fulfill 3 
    conditions:
    - It is defined in the file which name is stored in variable _LPB_FILE
    - Its name must begin with prefix "lpb_"
    - A comment \/* UNITS ... *\/ has be defined.

    See file of name saved in _LPB_FILE to get examples of implementation.

SEE ALSO: h_show
*/
{
  extern _LPB_FILE;
  
  // FIXME: this list should be returned by a function in LITpro world.
  reserved_args = ["ufreq", "vfreq", "wavelength", "bandwidth"];
  
  if (am_subroutine()) {
    // For collecting the SHORTDOC lines, function names and arguments.
    shortdocs = funcnames = arguments = [];
  } else {
    h = h_new();
  }
  
  f = open(_LPB_FILE);

  while (line= rdline(f)) {
    
    split= strtok(line);   // split = [<first word>, <remainder of line>]
    
    if (split(1)=="func") {
      // We found a function. Get its name.
      name= strtok(split(2), " \t(")(1);
      if (!strmatch(name, "lpb_")) continue
      // Prefix is ok. This function will be selected if UNITS comment is
      // found within the next 10 lines (without counting the lines of
      // blocks DOCUMENTS and SHORTDOC).
      doctext = units = vmin = vmax = [];
      shortdoc = "--- missing short doc ---";
      count = 10;        /* like help_worker function defined in std.i */
      
      // Collect the comments blocks.
      while ((line= rdline(f)) && count--) {
        if (!strlen(line)) break; // Empty line outside any block.
        if (strmatch(line, "/* DOCUMENT")) {
          do {
            grow, doctext, [line];
            if (strmatch(line, "*/")) break;
          } while (line= rdline(f));
          
        } else if (strmatch(line, "/* UNITS")) {
          do {
            grow, units, [line];
            if (strmatch(line, "*/")) { units = sum(units); break; }
          } while (line= rdline(f));
          
        } else if (strmatch(line, "/* VMIN")) {
          do {
            grow, vmin, [line];
            if (strmatch(line, "*/")) { vmin = sum(vmin); break; }
          } while (line= rdline(f));
          
        } else if (strmatch(line, "/* VMAX")) {
          do {
            grow, vmax, [line];
            if (strmatch(line, "*/")) { vmax = sum(vmax); break; }
          } while (line= rdline(f));
          
        } else if (strmatch(line, "/* SHORTDOC")) {
          // One single line in this case.
          shortdoc = strtrim(strpart(line,
                     strfind("/* SHORTDOC ",line)(2):strfind("*/",line)(1)));
        }
      }
      if (!is_void(units)) {
        // Thus the function is selected for being in the library.
        
        // Get the list of arguments as a vector of strings.
        args_string = strtok(sum(print(symbol_def(name))),"(,)");
        args=[];
        while ( (arg = (args_string = strtok(args_string(2),"(,)"))(1)) ) {
          // save in the list if the argument is not in the reserved list.
          if (!is_array(where(arg == reserved_args))) grow, args, arg;
        }

        if (am_subroutine()) {
          // Collect the lines.
          grow, shortdocs, shortdoc;
          grow, funcnames, name;
          grow, arguments,
                (numberof(args) == 1 ? args(1) : args(1)+sum(", "+args(2:)));
          
        } else {
          // Called as a function.
          // Open a new hash table for this function.
          key = strpart(name,5:0);     // Remove "lpb_" prefix.
          h_set, h, key, h_new();      // New entry for this function.
          hfunc = h_get(h, key);       // Reference on this entry.

          // Save help if it was found.
          if (!is_void(doctext)) help = sum(swrite(format="%s\n",doctext));
          else doctext = "";
          
          // Save and create sub-hash tables for each field to be documented,
          // UNITS, VMIN and VMAX.
          h_set, hfunc, fullname = name,
                        args = args,
                        help = doctext,
                        units = _lpb_functions_parser(units, "UNITS", args);
          if (vmin)
            h_set, hfunc, vmin = _lpb_functions_parser(vmin, "VMIN", args);
          if (vmax)
            h_set, hfunc, vmax = _lpb_functions_parser(vmax, "VMAX", args);
        }          
      }
    }
  }
  close, f;
  if (am_subroutine()) {
    funcnames += "()";
    write, format=swrite(format="    %%-%d",max(strlen(funcnames)))+"s - %s\n",
           funcnames, shortdocs;
  } else {
    return h;
  }
}



/*---------------------------------------------------------------------------*/
/*                             PUBLIC FUNCTIONS                              */
/*---------------------------------------------------------------------------*/

func lpb_punct(ufreq, vfreq, weight, x, y)
/* DOCUMENT lpb_punct(ufreq, vfreq, weight, x, y)
   
    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ)
    given in 1/rad, of a punctual object (Dirac function) at coordinates
    (X,Y) given in milliarcsecond.
    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    
    UFREQ and VFREQ must be conformable. The returned array is always
    complex and of dims dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Single point (Dirac function) */
/* UNITS weight, -, x, mas, y, mas */
/* VMIN  weight, 0, x,   -, y,   - */
/* VMAX  weight, 1, x,   -, y,   - */
{
  return weight * _lpb_shift(ufreq, vfreq, x, y);
}



/*---------------------------------------------------------------------------*/
/*                            CIRCULAR FUNCTIONS                             */
/*---------------------------------------------------------------------------*/


func lpb_disk(ufreq, vfreq, weight, x, y, radius)
/* DOCUMENT lpb_disk(ufreq, vfreq, weight, x, y, radius)
   
    Returns the Fourier transform, computed at spatial frequencies
    (UFREQ,VFREQ) given in 1/rad, of a normalized uniform disk of radius
    RADIUS (milliarcsecond) and centered at coordinates (X,Y)
    (milliarcsecond).

    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if RADIUS is negative, so it is advised
    to set a VMIN bound to zero or more on this parameter.
    
    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Uniform disk */
/* UNITS weight, -, x, mas, y, mas, radius, mas */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   - */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_MIN;
  if (radius < 0)
    error, "radius < 0 not allowed for disk: vmin bound could be set to zero";
  
  // Adding _LPB_DBL_MIN is a turn around for the singularity at r=0.
  // r is numerically unchanged where r > 2.00417e-292
  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_MIN;

  // Multiplying by _lpb_shift at the end ensures complex type.
  return (bessj1(2.*r)/r) * weight * _lpb_shift(ufreq,vfreq,x,y);
}


func lpb_circle(ufreq, vfreq, weight, x, y, radius)
/* DOCUMENT lpb_circle(ufreq, vfreq, weight, x, y, radius)
   
    Returns the Fourier transform, computed at spatial frequencies
    (UFREQ,VFREQ) given in 1/rad, of a normalized circle of radius RADIUS
    (milliarcsecond) and centered at coordinates (X,Y) (milliarcsecond).

    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if RADIUS is negative, so it is advised
    to set a vmin bound to zero or more on this parameter.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Circle */
/* UNITS weight, -, x, mas, y, mas, radius, mas */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   - */
{
  extern _LPB_PI, _LPB_MAS2RAD;
  if (radius < 0)
    error, "radius < 0 not allowed for circle: vmin bound could be set to zero";

  r = (_LPB_MAS2RAD * _LPB_PI * radius) * abs(ufreq, vfreq);
  
  // Multiplying by _lpb_shift at the end ensures complex type.
  return bessj0(2.*r)  * weight * _lpb_shift(ufreq, vfreq, x, y);
}


func lpb_gaussian(ufreq, vfreq, weight, x, y, fwhm)
/* DOCUMENT lpb_gaussian(ufreq, vfreq, weight, x, y, fwhm)
   
    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ),
    of a normalized gaussian with given FWHM centered at (X,Y) coordinates.
    
    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if FWHM is negative, so it is advised
    to set a vmin bound to zero or more on this parameter.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Gaussian */
/* UNITS weight, -, x, mas, y, mas, fwhm, mas */
/* VMIN  weight, 0, x,   -, y,   -, fwhm,   0 */
/* VMAX  weight, 1, x,   -, y,   -, fwhm,   - */
{
  extern _LPB_PI, _LPB_MAS2RAD;
  if (fwhm < 0) 
    error, "FWHM < 0 not allowed for gaussian: vmin bound could be set to zero";
  
  f = _LPB_MAS2RAD *_LPB_PI * fwhm;
  
  // Multiplying by _lpb_shift at the end ensures complex type.
  return  exp( -(f*f)/4./log(2.) * (ufreq*ufreq + vfreq*vfreq)  ) \
          * weight * _lpb_shift(ufreq, vfreq, x, y);
}


func lpb_ring(ufreq, vfreq, weight, x, y, radius, width)
/* DOCUMENT lpb_ring(ufreq, vfreq, weight, x, y, radius, width)
   
    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ),
    of a normalized uniform ring with internal radius RADIUS and external
    radius RADIUS+WIDTH, centered at (X,Y) coordinates.
    
    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if RADIUS or WIDTH is negative, so it is
    advised to set a vmin bound to zero or more on these parameters.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Uniform ring */
/* UNITS weight, -, x, mas, y, mas, radius, mas, width, mas */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0, width,   0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   -, width,   - */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_EPS, _LPB_DBL_MIN;
  if (radius < 0)
    error, "radius < 0 not allowed for ring: vmin bound could be set to zero";
  if (width < 0)
    error, "width < 0 not allowed for ring: vmin bound could be set to zero";

  if (width == 0.) {
    // Returns the Fourier transform of infinitely thin ring, i.e. a circle.
    return lpb_circle(ufreq, vfreq, weight, x, y, radius);
  }
  if (radius == 0.) {
    // Returns the Fourier transform of a disk of radius width.
    return lpb_disk(ufreq, vfreq, weight, x, y, width);
  }
  // This writting ensures that the result is normalized where
  // (ufreq,vfreq)=(0,0). Numerator and denominator of the expressions have
  // been simplified by radius (so we had to test radius == 0 before).
  alpha = 1. + width/double(radius);
  // r is numerically unchanged where r > 2.00417e-292
  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_MIN;
  /*
  return (alpha * bessj1(2.*alpha*r) - bessj1(2.*r)) / ((alpha*alpha-1.)*r) \
    * weight * _lpb_shift(ufreq,vfreq,x,y);
  this writing induces on MacIsa a floating point interrupt 
  */

  return ((alpha * bessj1(2.*alpha*r)/r) - (bessj1(2.*r)/ r))           \
    * (weight/(alpha*alpha-1.))   * _lpb_shift(ufreq,vfreq,x,y);

}



/*---------------------------------------------------------------------------*/
/*                            ELONGATED FUNCTIONS                            */
/*---------------------------------------------------------------------------*/

func lpb_elong_disk(ufreq, vfreq, weight, x, y,
                    minor_radius, axis_ratio, orientation)
/* DOCUMENT lpb_elong_disk(ufreq, vfreq, weight, x, y,
                           minor_radius, axis_ratio, orientation)

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ), of
    a normalized ellipse centered at coordinates (X,Y), with a ratio AXIS_RATIO
    between the minor radius MINOR_RADIUS and the major one,
    turned from the horizontal axis with angle ORIENTATION, in degrees.
    
    AXIS_RATIO = major_axis / minor_axis

    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if MINOR_RADIUS is negative or if AXIS_RATIO
    is smaller than 1, so it is advised to set a vmin bounds on these
    parameters.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions
*/
/* SHORTDOC Ellipse (elongated disk) */
/* UNITS weight, -, x, mas, y, mas, minor_radius, mas, axis_ratio, -,
   orientation, degrees */
/* VMIN  weight, 0, x,   -, y,   -, minor_radius,   0, axis_ratio, 1,
   orientation, -90 */
/* VMAX  weight, 1, x,   -, y,   -, minor_radius,   -, axis_ratio, -,
   orientation,  90 */
{
  if (minor_radius < 0)
    error, "minor_radius < 0 not allowed for ellipse: "+
            "vmin bound could be set to zero";
  if (axis_ratio < 1)
    error, "AXIS_RATIO < 1 not allowed for ellipse: "+
           "vmin bound could be set to 1";

  T_uv =_lpb_transform(ufreq, vfreq, double(axis_ratio), 1., orientation);
  return lpb_disk(T_uv(..,1), T_uv(..,2), weight, 0., 0., minor_radius)\
         * _lpb_shift(ufreq, vfreq, x, y);
}


func lpb_elong_gaussian(ufreq, vfreq, weight, x, y,
                        minor_fwhm, fwhm_ratio, orientation)
/* DOCUMENT lpb_elong_gaussian(ufreq, vfreq, weight, x, y,
                               minor_fwhm, fwhm_ratio, orientation)

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ),
    of a normalized elongated gaussian centered at (X,Y) coordinates. The
    sizes of the function in two orthogonal directions are given by the
    narrowest FWHM (MINOR_FWHM) and by the ratio FWHM_RATIO between the major
    FWHM and MINOR_FWHM, in the same way as for an ellipse:
    
    FWHM_RATIO = major_fwhm / MINOR_FWHM.
    
    ORIENTATION is measured in degrees.
    
    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    The function returns an error if FWHM_RATIO is smaller than 1, so it is
    advised to set a vmin bound to 1 on this parameter.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions, lpb_gaussian, lpb_elong_disk
*/
/* SHORTDOC Elongated Gaussian */
/* UNITS weight, -, x, mas, y, mas, minor_fwhm, mas, fwhm_ratio, -,
   orientation, degrees */
/* VMIN  weight, 0, x,   -, y,   -, minor_fwhm,   0, fwhm_ratio, 1,
   orientation, -90 */
/* VMAX  weight, 1, x,   -, y,   -, minor_fwhm,   -, fwhm_ratio, -,
   orientation, 90 */
{
  if (minor_fwhm < 0)
    error, "minor_fwhm < 0 not allowed for elongated gaussian: "+
           "vmin bound could be set to zero";
  if (fwhm_ratio < 1)
    error, "FWHM_RATIO < 1 not allowed for elongated gaussian: "+
           "vmin bound could be set to 1";

  T_uv = _lpb_transform(ufreq, vfreq, double(fwhm_ratio), 1., orientation);
  return lpb_gaussian(T_uv(..,1),T_uv(..,2),weight,0,0,minor_fwhm) \
         * _lpb_shift(ufreq, vfreq, x, y);
}


func lpb_elong_ring(ufreq, vfreq, weight, x, y,
                    minor_radius, radius_ratio, width, orientation)
/* DOCUMENT lpb_elong_ring(ufreq, vfreq, weight, x, y,
                           minor_radius, radius_ratio, width, orientation)

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ),
    of a normalized uniform elongated ring centered at (X,Y) coordinates. 
    The sizes of the function in two orthogonal directions are given
    by the narrowest internal radius (MINOR_RADIUS) and by the ratio
    RADIUS_RATIO between the widest internal radius and MINOR_RADIUS, in
    the same way as for an ellipse:
    
    RADIUS_RATIO = major_radius / MINOR_RADIUS.
    
    In the direction of MINOR_RADIUS, the external radius is
    MINOR_RADIUS + WIDTH. In the direction of major_radius, the width is
    magnified by the ratio RADIUS_RATIO, so that the external radius is
    major_radius + RADIUS_RATIO * WIDTH.
    
    ORIENTATION is measured in degrees.

    WEIGHT is the intensity coefficient. WEIGHT=1 means total energy is 1.
    
    The function returns an error if MINOR_RADIUS is negative or if
    RADIUS_RATIO is smaller than 1, so it is advised to set a vmin bounds
    on these parameters.

    UFREQ and VFREQ must be conformable. The returned array is always
    complex and with dimensions dimsof(UFREQ,VFREQ).

SEE ALSO: lpb_functions, lpb_ring
*/
/* SHORTDOC Elongated uniform ring */
/* UNITS weight, -, x, mas, y, mas, minor_radius, mas, radius_ratio, -,
   width, mas, orientation, degrees */
/* VMIN  weight, 0, x,   -, y,   -, minor_radius,   0, radius_ratio, 1,
   width,   0, orientation, -90 */
/* VMAX  weight, 1, x,   -, y,   -, minor_radius,   -, radius_ratio, -,
   width,   -, orientation,  90 */
{
  if (minor_radius < 0)
    error, "minor_radius < 0 not allowed for elongated ring: "+
            "vmin bound could be set to zero";
  if (radius_ratio < 1)
    error, "radius_ratio < 1 not allowed for elongated ring: "+
           "vmin bound could be set to 1";
    
  T_uv = _lpb_transform(ufreq, vfreq, double(radius_ratio), 1., orientation);
  return lpb_ring(T_uv(..,1), T_uv(..,2), weight, 0,0, minor_radius, width) \
         * _lpb_shift(ufreq, vfreq, x, y) ;
}



/*---------------------------------------------------------------------------*/
/*                 Center TO LIMB DARKENING FUNCTIONS                        */
/*---------------------------------------------------------------------------*/
/* require "yeti_gsl.i"
 */

func lpb_limb_power(ufreq, vfreq, weight, x, y, radius, power)  
/* DOCUMENT lpb_limb_power(ufreq, vfreq, weight, x, y, radius, power)

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ), of
    a center-to-limb darkened disk of radius RADIUS, centered at coordinates (X,Y).
    The brightness distribution o, if expressed versus mu, which is the cosine
    of the azimuth of a surface element of the star,  follows
    a law of power POWER, and is normalized for mu = 1 (center of the star).
    o(mu) = mu^POWER.
    WEIGHT is the intensity coefficient.
    NB: also works if RADIUS=0 or abs(ufreq,vfreq)=0.  

SEE ALSO: lpb_functions
*/
/* SHORTDOC Limb-darkened disk with power law */
/* UNITS weight, -, x, mas, y, mas, radius, mas, power, - */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0, power, 0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   -, power, - */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_EPS ;
  if (power <0 )                                                                    
      error, "power < 0 not allowed for this limb-darkening function: "+
           "vmin bound could be set to 0";

  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_EPS;
  
  k = power/2. +1.;
  return k * gsl_sf_gamma(k) * (gsl_sf_bessel_Jnu(k, 2. *r) /r^k) \
            * weight  * _lpb_shift(ufreq, vfreq, x, y) ;
}

func lpb_limb_linear(ufreq, vfreq, weight, x, y, radius, a1_coeff)  
/* DOCUMENT lpb_limb_linear(ufreq, vfreq, weight, x, y, radius, a1_coeff)

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ), of
    a center-to-limb darkened disk of radius RADIUS, centered at
    coordinates (X,Y).  The brightness distribution o, if expressed versus
    mu, the cosine of the azimuth of a surface element of the star, follows
    a linear law of coefficient A1_COEFF ( ]0,1[ ), and is normalized for mu = 1
    (center of the star).
    o(mu) = 1 -A1_COEFF(1-mu).
    WEIGHT is the intensity coefficient.
    NB: also works if RADIUS=0 or abs(ufreq,vfreq)=0.

SEE ALSO: lpb_functions
*/
/* SHORTDOC Limb-darkened disk with linear law */
/* UNITS weight, -, x, mas, y, mas, radius, mas, a1_coeff, - */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0, a1_coeff, 0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   -, a1_coeff, 1 */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_EPS ;
  //  if ((a1_coeff > 1) || (a1_coeff <0) )
   if (a1_coeff <0 )                                                                    
      error, "a1_coeff < 0 not allowed for this limb-darkening function: "+
           "vmin bound could be set to 0";
   
  if (a1_coeff >1 ) 
      error, "a1_coeff > 1 not allowed for this limb-darkening function: "+
           "vmax bound could be set to 1";
  
  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_EPS;
   
  N_fact = 1.- a1_coeff/3.;
  return ((1-a1_coeff)* bessj1(2.*r)/r \
         + (a1_coeff*sqrt(_LPB_PI)/2. * gsl_sf_bessel_Jnu(1.5, 2.*r)/r^(1.5))) \
         * weight /N_fact * _lpb_shift(ufreq, vfreq, x, y) ;
}

func lpb_limb_quadratic(ufreq, vfreq, weight, x, y, radius, a1_coeff, a2_coeff)  
/* DOCUMENT lpb_limb_quadratic(ufreq, vfreq, weight, x, y, radius, a1_coeff, a2_coeff)  

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ), of
    a center-to-limb darkened disk of radius RADIUS, centered at
    coordinates (X,Y).  The brightness distribution o, if expressed versus
    mu, the cosine of the azimuth of a surface element of the star, follows
    a quadratic law of coefficients A1_COEFF, A2_COEFF ( ]0,1[ ), and is normalized
    for mu = 1 (center of the star).
    o(mu) = 1 -A1_COEFF(1-mu) - A2_COEFF(1-mu)^2.
    WEIGHT is the intensity coefficient.
    NB: also works if RADIUS=0 or abs(ufreq,vfreq)=0.

SEE ALSO: lpb_functions
 */
/* SHORTDOC Limb-darkened disk with quadratic law */
/* UNITS weight, -, x, mas, y, mas, radius, mas, a1_coeff, -, a2_coeff, - */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0, a1_coeff, 0, a2_coeff, 0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   -, a1_coeff, -, a2_coeff, - */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_EPS ;
  if ((a1_coeff > 1) || (a1_coeff <0) )  \
    error, "a1_coeff < 0 or >1 not allowed for this limb-darkening function:"+
       "vmin bound could be set to 0 and vmax bound could be set to 1";
  if ((a2_coeff > 1) || (a2_coeff <0) )  \
    error, "a2_coeff < 0 or >1 not allowed for this limb-darkening function:"+
       "vmin bound could be set to 0 and vmax bound could be set to 1";

  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_EPS;
  N_fact = 1.- a1_coeff/3.- a2_coeff/6.;
  
  return  ((1-a1_coeff-a2_coeff)* bessj1(2.*r)/r \
    + ((a1_coeff+2.*a2_coeff)*sqrt(_LPB_PI)/2. * gsl_sf_bessel_Jnu(1.5, 2.*r)/r^(1.5))\
    - a2_coeff * gsl_sf_bessel_Jn(2, 2.*r)/(r*r))\
    * weight /N_fact * _lpb_shift(ufreq, vfreq, x, y) ;
}

func lpb_limb_sqrt(ufreq, vfreq, weight, x, y, radius, a1_coeff, a2_coeff)  
/* DOCUMENT lpb_limb_sqrt(ufreq, vfreq, weight, x, y, radius, a1_coeff, a2_coeff)  

    Returns the Fourier transform, at spatial frequencies (UFREQ,VFREQ), of
    a center-to-limb darkened disk of radius RADIUS, centered at
    coordinates (X,Y).  The brightness distribution o, if expressed versus
    mu, the cosine of the azimuth of a surface element of the star, follows
    a square root law of coefficients A1_COEFF, A2_COEFF ( ]0,1[ ), and is normalized
    for mu = 1 (center of the star).
    o(mu) = 1 -A1_COEFF(1-mu) - A2_COEFF(1-sqrt(mu)).
    WEIGHT is the intensity coefficient.
    NB: also works if RADIUS=0 or abs(ufreq,vfreq)=0.

SEE ALSO: lpb_functions
*/
/* SHORTDOC Limb-darkened disk with square root law */
/* UNITS weight, -, x, mas, y, mas, radius, mas, a1_coeff, -, a2_coeff, - */
/* VMIN  weight, 0, x,   -, y,   -, radius,   0, a1_coeff, 0, a2_coeff, 0 */
/* VMAX  weight, 1, x,   -, y,   -, radius,   -, a1_coeff, -, a2_coeff, - */
{
  extern _LPB_PI, _LPB_MAS2RAD, _LPB_DBL_EPS ;
   if ((a1_coeff > 1) || (a1_coeff <0) )  \
    error, "a1_coeff < 0 or >1 not allowed for this limb-darkening function:"+
       "vmin bound could be set to 0 and vmax bound could be set to 1";
  if ((a2_coeff > 1) || (a2_coeff <0) )  \
    error, "a2_coeff < 0 or >1 not allowed for this limb-darkening function:"+
       "vmin bound could be set to 0 and vmax bound could be set to 1";

  r = (_LPB_MAS2RAD *_LPB_PI * radius) * abs(ufreq, vfreq) + _LPB_DBL_EPS;
  N_fact = 1.- a1_coeff/3.- a2_coeff/5.;
  
  return  ((1-a1_coeff-a2_coeff)* bessj1(2.*r)/r \
    + (a1_coeff*sqrt(_LPB_PI)/2. * gsl_sf_bessel_Jnu(1.5, 2.*r)/r^(1.5)) \
    - a2_coeff * gsl_sf_gamma(1.25) *gsl_sf_bessel_Jnu(1.25, 2.*r)/r^(1.25))\
    * weight /N_fact * _lpb_shift(ufreq, vfreq, x, y) ;
}



/*---------------------------------------------------------------------------*/
/*                            PRIVATE FUNCTIONS                              */
/*---------------------------------------------------------------------------*/

func _lpb_functions_parser(field_string, label, args)
/* DOCUMENT _lpb_functions_parser(field_string, label, args)
    
    Private function used by lpb_functions. FIELD_STRING is a string collected
    in file which name is stored in _LPB_FILE. FIELD_STRING are commented
    strings like:
      "UNITS weight, -, x, mas, y, mas"
      "VMIN  weight, 0, x,   -, y,   -"
    or more generally:
      "LABEL <arg>, <value>, <arg>, <value>, <arg>, <value> ...
      
    LABEL is the upercase string identifying FIELD_STRING. In the above
    examples, labels are "UNITS" and "VMIN".
    ARGS is an array of string which gives the names of the arguments whose
    values are to be looked for in FIELD_STRING.
    
    The function parse FIELD_STRING string and returns a hash table giving
    the corresponding value for each argument. In the above example we would
    get for UNITS:
    
    TOP (hash_table, 4 entries)
    |- weight (string) ""
    |- x (string) "mas"
    `- y (string) "mas"

SEE ALSO: lpb_functions
*/
{

  // Ensure name exists. name should be defined in the caller (i.e
  // lpb_functions()). This ensure name is defined when testing directly
  // parsing function from the command line.
  if (is_void(name)) name="";
  
  // Get the contents of field_string as a vector of string
  // ["<arg>", "<value>", "<arg>", "<value>", ...]
  // First remove the label and separate the first token.
  field_string = strtok(strtrim(strpart(field_string,
      strfind(label, field_string)(2)+1:strfind("*/",field_string)(1))),", ");
      
  // Now break down field_string into a vector of tokens (strings).
  tokens = field_string(1);
  while ( (token = (field_string = strtok(field_string(2),", "))(1)) )
    grow, tokens, token;

  // Prepare the hash table to be returned
  h = h_new();
  
  // Save in the hash table the value corresponding to each argument,
  // after checking various errors.
  for (i=1; i<=numberof(args) ; ++i) {
    arg = args(i);
    index = where(tokens == arg);   // find this arg in the list of tokens.
    if (!is_array(index)) {
      // This is a programmer bug.
      error, "argument \""+arg+"\" is missing in "+label+ " of "+
             name+"()";
    }
    if (numberof(index) > 1) {
      // This is a programmer bug:
      // maybe 2 different values for the same arguments.
      error, "argument \""+arg+"\" is defined several times in "+
             label+" of "+name+"()";
    }
    value = tokens(index(1)+1);     // get the corresponding value.
    if ( value == "-" ) value = ""; // No value for this arg.
    h_set, h, arg, value;           // Save in hash table.
  }
  return h;
}


func _lpb_shift(ufreq, vfreq, x, y)
/* DOCUMENT _lpb_shift(ufreq, vfreq, x, y)
   
    Returns complex factor to apply in the Fourier transform at frequencies
    (UFREQ,VFREQ) to account for a shift (X,Y) in image space.
    X, Y are given in milliarcseconds.
SEE ALSO:
*/
{
  extern _LPB_MAS2RAD,_LPB_PI;
  phase = 2. * _LPB_PI * _LPB_MAS2RAD * (x*ufreq + y*vfreq);
  return cos(phase) - 1i*sin(phase);
}


func _lpb_transform(ufreq, vfreq, t_ana, t_homo, rotation)
/* DOCUMENT _lpb_transform(ufreq, vfreq, t_ana, t_homo, rotation)
   
    Returns the new spatial frequencies when the object has got geometrical
    transformations, successively a rotation,  an anamorphose and a homothetie.
    (u,v)--> Transpose(Inverse(T))(u,v), with matrix T = HAR;
    Inverse(R)= |cos(rotation) -sin(rotation)|
                |sin(rotation)  cos(rotation)|   rotation angle in degrees
    Inverse(A)= |t_ana 0|
                |0     1|  t_ana = ratio of anamorphose, >0
    Inverse(H)= |t_homo   0|
                |0   t_homo|  t_homo = ratio of homothetie >0
SEE ALSO:
*/
{
  extern _LPB_DEG2RAD;
  rotation *= _LPB_DEG2RAD;
  _ufreq =  ufreq * (cos(rotation)* t_ana* t_homo) + \
            vfreq * (sin(rotation)* t_ana* t_homo);
  _vfreq = -ufreq * (sin(rotation)* t_homo) + vfreq * (cos(rotation)* t_homo);
  return [_ufreq,_vfreq];
}


func _lpb_test_functions(void)
{
  extern _LPB_DBL_EPS, _LPB_DBL_MIN
  // Test if the current machine complies with IEEE standard for floating
  // values.
  if (_LPB_DBL_MIN/2 != 0) {
    IEEE_compliant = 1;
    write, format="%s\n",
           "This machine complies with IEEE standard for floating values.";
    
  } else {
    IEEE_compliant = 0;
    lp_warn,
      "This machine does not comply with IEEE standard for floating values.",
      "If a floating value is less than xmin, many computers underflow its",
      "value to zero. The IEEE standard specifies a more graceful kind of",
      "underflow: As a value becomes smaller than xmin, its exponent is",
      "frozen at the smallest allowed value, while its mantissa is decreased,",
      "acquiring leading zeros and \"gracefully\" losing precision.\n";
  }
  write, format="_LPB_DBL_MIN is: %g\n", _LPB_DBL_MIN;
  write, format="_LPB_DBL_EPS is: %g\n", _LPB_DBL_EPS;
  
  // Bessel functions
  _lpb_test_function, "bessj1",
    //       answer,              x
    [[ _LPB_DBL_MIN, 2*_LPB_DBL_MIN],
     [ _LPB_DBL_EPS, 2*_LPB_DBL_EPS]];

  // lpb_disk
  _lpb_test_function, "lpb_disk",
    // answer, ufreq, vfreq, weight,   x,   y, radius
    [[     1.,    0.,    0.,     1.,  0.,  0.,     1.],
     [     1.,    0.,    0.,     1.,  0.,  0.,     0.],
     [     1.,   1e9,   1e9,     1.,  0.,  0.,     0.]],
     "         ufreq, vfreq, weight, x, y, radius";

  // lpb_circle
  _lpb_test_function, "lpb_circle",
  [[     1.,    0.,    0.,     1.,  0.,  0.,     1.],
   [     1.,    0.,    0.,     1.,  0.,  0.,     0.],
   [     1.,   1e9,   1e9,     1.,  0.,  0.,     0.]],
   "         ufreq, vfreq, weight, x, y, radius";

  // lpb_gaussian
  _lpb_test_function, "lpb_gaussian",
    // anwser, ufreq, vfreq, weight,   x,   y,  fwhm
    [[     1.,    0.,    0.,     1.,  0.,  0.,     0.],
     [     1.,   1e9,   1e9,     1.,  0.,  0.,     0.]],
     "         ufreq, vfreq, weight, x, y, fwhm";

  // lpb_ring
  _lpb_test_function, "lpb_ring",
    // anwser, ufreq, vfreq, weight,   x,   y, radius, width
   [[     1.,    0.,    0.,     1.,  0.,  0.,     1.,    1.],
     [     1.,    0.,    0.,     1.,  0.,  0.,    1e6,  1e-6],
     [     1.,    0.,    0.,     1.,  0.,  0.,     1.,  1e-6],
     [     1.,    0.,    0.,     1.,  0.,  0.,     0.,    1.],
     [     1.,    0.,    0.,     1.,  0.,  0.,     1.,    0.],
     [     1.,   1e9,   1e9,     1.,  0.,  0.,     0.,    0.]],
     "         ufreq, vfreq, weight, x, y, radius, width";
}


func _lpb_test_function(function_name, arguments, comments)
{
  local display;
  
  // Arguments must be double (not complex)
  if (is_complex(arguments)) error, "arguments cannot be complex";
  arguments = double(arguments);
  
  // Sparse the arguments
  d = dimsof(arguments);
  if (d(1) == 1) {
    nb_tests = 1;    // arguments is a single vector.
  } else if (d(1) == 2) {
    nb_tests = d(3); // arguments is a set of vectors.
  } else {
    error, "ARGUMENTS has more than 2 dimensions !";
  }
  // We must have at least 2 arguments: answer, parameter.
  // Generally more than 4: answer, ufreq, vfreq, parameter
  nb_args = d(2);
  if (nb_args < 2) error, "Less than 2 arguments per test"
  
  write, format="\n--------- test of %s ---> diff, relative%%\n", function_name;
  write, format="%s\n", comments;

  // Check that function exists.
  if (is_void(function_name)) error, "function name is void";
  if (!symbol_exists(function_name) 
      || !is_func( (function=symbol_def(function_name)) )) {
        write, format="\tWARNING: function %s not found\n", function_name;
        return;
  }

  for (it = 1; it <= nb_tests; it++) {
  
    // Exec function, depending on the number of arguments.
    if (nb_args == 2) {
      r = function(arguments(2, it));

    } else if (nb_args == 3) {
      r = function(arguments(2, it), arguments(3, it));

    } else if (nb_args == 4) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it));
      
    } else if (nb_args == 5) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it),
                   arguments(5, it));
                   
    } else if (nb_args == 6) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it),
                   arguments(5, it), arguments(6, it));
                   
    } else if (nb_args == 7) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it),
                   arguments(5, it), arguments(6, it), arguments(7, it));

    } else if (nb_args == 8) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it),
                   arguments(5, it), arguments(6, it), arguments(7, it),
                   arguments(8, it));

    } else if (nb_args == 9) {
      r = function(arguments(2, it), arguments(3, it), arguments(4, it),
                   arguments(5, it), arguments(6, it), arguments(7, it),
                   arguments(8, it), arguments(9, it));
    } else {
      error, "more than 9 arguments not supported (yet)"
    }

    answer = arguments(1, it);
    diff = r - answer;
    if (is_complex(diff) && diff.im != 0) {     // r may be complex
      diff_string = swrite(format="%g%+gi", diff.re, diff.im);
    } else {
      diff_string = swrite(format="%g",double(diff));
    }
    if (answer != 0) {
      err = diff/answer*100.;
      if (is_complex(err) && diff.im != 0) {    // r may be complex
        err_string = swrite(format="%g%+gi%%",err.re,err.im);
      } else {
        err_string = swrite(format="%g%%",double(err));
      }
    } else {
      err_string = "";
    }    
    // Record for later display (at the end of the tests)
    line = [];
    if (nb_args > 2) {
      grow, line, swrite(format="%g,", arguments(2:-1, it));
    }
    grow, line, swrite(format="%g)", arguments(0, it)),
                " <-> ",
                swrite(format="%g", answer),
                " ---> ",
                diff_string,
                err_string;
    grow, display, [line];
  }
  _lp_print_table, display, indent=function_name+"(", no_trim=1;
}
